! 20.07.2004 
!+
! Module: fluxfit_mod
!
! Authors:  Christopher Brindle - MSSL
!           
!          
! Description: Fits synthetic spectra to source count-rates to obtain fluxes
! 
! Comments: 
! 
! $Id: module.f90,v 1.5 1998/01/13 08:39:15 gvacanti Exp $
!
! History: 
!       
!
!************************************************************************************************
module fluxfit_mod
  use param
  use cal
  use dal
  use Errorhandling
  use sla
  use getFilterID_mod
 
  use omallocation_mod
  use omutility_mod 
  use copytable_mod
  use omgraph_mod
  use leastsquares_mod
  
  implicit none 

  public


    type SourceData
        real(single) :: mag(7) = -1.0
        real(single) :: magErr(7) = -1.0
        real(single) :: counts(7) = -1.0
        real(single) :: countsErr(7) = -1.0
        real(single) :: weight(7) = -1.0
        real(single) :: flux(7) = -1.0
        real(single) :: fluxerr(7) = -1.0
        character(4) :: w(7)
        integer(int8) :: filterIndex
        character(6) :: spectralType = "------"
        real(single) :: fluxConstant = -1.0
        real(single) :: fluxConstantError = -1.0

        real(single) :: redchisq = -1.0
        integer(int8) :: noFilters
        character(4) :: filter(7)
      
    end type SourceData


    type FittedData
        real(single) :: rmsResidual
        integer :: templateNo
        real(single) :: constant
        real(single) :: constantError
        integer :: ndf
        real(single) :: redchisq = -1.0
        character(7) :: spectralType = "------"
        real(single) :: flux(7), fluxErr(7)
        real(single) :: residual(7) 

        real(single) :: fluxConvert(7) = (/5.242e-15, 1.754e-15, 4.169e-16, 1.776e-16, 1.041e-16, 2.178e-16, -1.0/)
 
    end type fittedData 
  

    real(double), dimension(1 : 7) :: modelData, fitData, fitDataErr, weight, wData 
  
    integer :: mFit
 

    type(SourceData), pointer :: magTable(:) => null()
    type(SourceData), pointer :: templateData(:) => null()
    logical(boolean) :: fitted
!    real(single) :: nullValue 
contains 


!********************************************************
!
!*******************************************************
subroutine GetZeroCounts(filter, f0)
    character(*), intent(in) :: filter(:)
    real(double), intent(Out) :: f0(:)
    integer :: i, nFilts, filterIdInt
    real(double) :: rate, mag, error, errorPlux, errorMinus
    
    nFilts = size(filter)
    rate = 10.0_double
    error = sqrt(rate)

    do i = 1, nFilts

        !  Get the filter Id integer value
        filterIdInt = filterIdToInt(trim(filter(i)))

        call CAL_setState(OM, filterId = filterIdInt)

        call CAL_omPhotoMagnitude(rate, mag, -error, error, errorPlux, errorMinus)

        f0(i) = rate * (10.0_double ** (0.4_double * mag)) 
        write(errstr, '(a, a, a, e12.4)') "Filter ", trim(filter(i)), " f0=", f0(i)
        call message(AppMsg, Verbose)
        
    end do

end subroutine GetZeroCounts



!********************************************************
!
!*******************************************************
subroutine GetFluxConstants(filter, f0)
    character(*), intent(in) :: filter(:)
    real(double), intent(Out) :: f0(:)
    integer :: i, nFilts
    character(100) :: string
    
    nFilts = size(filter)
  

    do i = 1, nFilts

        string = "FCF" // trim(filter(i))
        if (CAL_hasMiscellaneousData(trim(string))) then
           call CAL_getMiscellaneousData(trim(string), f0(i))
           write(*, *) "Filter ", filter(i), " flux constant=", f0(i)
        end if 
        
    end do

end subroutine GetFluxConstants


!***************************************************
! Load the stellar template data
! 7 wavebands
! UVW2, UVM2, UVW1, U, B, V and WHITE
!***************************************************
subroutine LoadTemplateData(templatedata)
    type(sourceData), pointer :: templateData(:)
    character(200) :: fileName(3) = &
          (/"/home/ssco2/sasdevel/packages/omsrclistcomb1/data/res_counts_om_fff07new_numb.txt", &
            "/home/ssco2/sasdevel/packages/omsrclistcomb1/data/res_flux_om_fff07new_numb.txt  ", &
            "/home/ssco2/sasdevel/packages/omsrclistcomb1/data/res_flux_om_fff07new_text.txt  "/)
 


    character(11) :: text
    integer :: i, k, no 
    real(single) :: counts(7)

    write(errstr, '(a)') "Loading stellar template data"
    call message(AppMsg, Verbose)
    
    open(1, status = "OLD", err = 1, file = trim(fileName(1)))


    do i = 1, 7
        read(1, *) text

    end do

    ! Find the number of records in the file
    k = 0
    do 
        read(1, *, end = 2) counts
        k = k + 1
    end do

2   close(1)

    ! Load the data
    open(1, status = "OLD", err = 1, file = trim(fileName(1)))  ! Counts
    open(2, status = "OLD", err = 1, file = trim(fileName(2)))  ! Fluxes 
    open(3, status = "OLD", err = 1, file = trim(fileName(3)))  ! Spectral types 


    do i = 1, 7
        read(1, *) text
        read(2, *) text
        read(3, *) text
    end do
   
    allocate(templateData(1 : k))

    k = 0

    do 
        read(1, *, end = 3) no, counts
 
        k = k + 1
        
        templateData(k)%counts = counts

        read(2, *, end = 3) no, counts 

        ! Convert from 0 to 20 magnitude
        counts = counts * 10.0 ** (-0.4 * 20.0)        
        templateData(k)%flux = counts

        read(3, '(a11)', end = 3) text
        text = adjustl(text)
        templateData(k)%spectralType = text(1 : 6) 
!if (k .lt. 20) write(*, *) "Line 175 spectral type=",  templateData(k)%spectralType
    end do

3   close(1)
    close(2)
    close(3)
    return

1   write(errstr, *) "Error opening file ", trim(fileName(1))    
    call fatal("File error", trim(errstr))

end subroutine LoadTemplateData

!****************************************************
! Tries to fit each template spectrum to the given
! source-data
!
!****************************************************
subroutine FitSpectra(sourceNo, magList, templateData, fitted, fit)
    integer, intent(in) :: sourceNo
    type(sourceData), intent(inout) :: magList
    type(sourceData), pointer :: templateData(:)
    logical(boolean), intent(out) :: fitted
    type(FittedData), intent(out) ::fit
  
 
    integer(int16) :: templateNo
 
    real(double) :: x(1)
    real(single) :: residual(7)
    integer :: i, m, n
    real(double) ::  xMin(1), redchisq, redchisqMin
    logical(boolean) :: use(7)
    real(single), dimension(1 : 7)  :: wavelength = (/0.5, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0/) 
    
    integer :: j
    use= .false.

    m = 0
    fitData = 0.0_double
    magList%flux = nullValue
    magList%fluxErr = nullValue
    magList%redchisq = nullValue
    fit%flux = nullValue
    fit%fluxerr = nullValue
    fitted = .false.
    weight = 0.0_double

!if (sourceNo .gt. 10) return  



    
    do j = 1, 7

        ! Only use the OM data where magnitudes have been computed 
        if (magList%counts(j) .gt. 0.0_double .and. magList%counts(j) .lt. 1000.0) then
            use(j) = .true.                     
            m = m + 1
            fitData(m) = magList%counts(j)
                  
            fitDataErr(m) = magList%countsErr(j)
            weight(m) = 1.0 / (fitDataErr(m) * fitDataErr(m))
        end if 
    end do
    if (m .ne. 7) fitData(m + 1 : 7) = 0.0_double


!write(*, *) "Line 276 fit spectra fitData=", fitData
!write(*, *) "Line 276 fit spectra fitDataerr=", fitDataErr
!write(*, *) "m=", m

    if (m .lt. 3) then  ! Not enough data- minimum 4 points
        fitted = .false.

        return


    end if

if (m .lt. 2) then
write(*, *) "m=", m
call fatal("bug at line 302")
end if





!write(*, *) "Line 314"
    !*********************************************************************
    ! Loop through all the template data to find the best match with
    ! the OM data
    !********************************************************************
    n = 1
    redchisQMin = 1.0e20    
    fitted = .true.






    
    do i = 1, ubound(templateData, 1)

!write(*, *) "line 326 fitting to template ", i
        modelData = templatedata(i)%counts ! Initialise model data to the template data
        m = 0
       
        do j = 1, 7
            if (use(j) .and. modelData(j) .ge. 0.0) then                     
                m = m + 1
                fitData(m) = magList%counts(j)
                fitDataErr(m) = magList%countsErr(j)
                weight(m) = 1.0 / (fitDataErr(m) * fitDataErr(m))


                modelData(m) = modelData(j)    
                wData(m) = wavelength(j)
            else
                use(j) = .false.
            end if
               
           
        end do


if (m .lt. 2) then
write(*, *) "m=", m
call fatal("bug at line 350")
end if




!m = 6
!do j = 1, 6
!modelData(j) = real(1000000 * j - 100000 * j * j)
!fitData(j) = real(1000 * j - 100 * j * j)
!weight(j) = 1.0
!end do

        if (m .ne. 7) modelData(m + 1 : 7) = 0.0_double

!write(*, *) "Line 358 fitting spectrum"
!write(*, *) "fitdata=", fitData

!write(*, *) "weight=", weight
!write(*, *) "modeldata=", modelData

        !*************************************************************
        ! Try to fit the template spectrum to the data
        !************************************************************
if (m .lt. 2) then
write(*, *) "m=", m
call fatal("bug at line 355")
end if


        if ( FitSpectrum(x, m, redChiSq, residual)) then

!write(*, *) "line 362 redchisq=", redchisq
            if (redChisq .lt. redChiSqMin) then ! Store the new minimum chi-squared fit values
                redchisQMin = redChiSq
                templateNo = i
                xMin(1) = x(1)

            end if   
        end if
    end do
 
    !**********************************************************
    ! A fit to the data is deemed successful if the
    ! minimum chi-squared fit is <= 1.0
    !********************************************************* 
    if (redchisqMin .gt. 1.0) then
        fitted = .false.


        magList%redchisq = nullValue
        magList%spectralType = "NO FIT" 

    else ! Acceptable fit- one with minimum chi-squared


        mfit = m

        m = 0
        do j = 1, 7
            if (use(j)) then
                m = m + 1
                modelData(m) = templatedata(tempLateNO)%counts(j)
            end if
        end do  

        fit%redChisQ = redChisqMin
        fit%templateNo = templateNo
        fit%constant = xmin(1) 
        fit%spectralType = templatedata(tempLateNO)%spectralType
        fit%ndf = m - 1          
        fit%residual = residual
         
        ! Error in fit constant


if (m .lt. 2) then
write(*, *) "m=", m
call fatal("Bug at line 400 in fluxfit_mod")
end if
        
        fit%constantError = ComputeFitConstantError(xmin(1), m)


        write(errstr, '(a, i6, a, e12.4, a, e12.4, a, e12.4, a, i3, a, a)') &
             " Spectrum successfully fitted to source", sourceNo, " x= ", & 
              xmin(1),  "+/-,", fit%constantError, &
             " Reduced chi-squared=",  redChisqMin, " template no=", tempLateNO, &
             " spectral type=", fit%spectralType

        call message(AppMsg, Verbose)

        do j = 1, 7
                               
                            
            fit%flux(j) = templatedata(tempLateNo)%flux(j) * xmin(1)
            if (magList%countsErr(j) .ne. nullValue .and. magList%countsErr(j) .gt. 0.0) then 
                fit%fluxErr(j) = fit%flux(j) * fit%constantError / fit%constant


            else
                fit%fluxErr(j) = nullValue
            end if
            

        end do

      
        magList%redchisq = fit%redchisq
        
        magList%spectralType = trim(templatedata(tempLateNO)%spectralType)
        magList%fluxConstant = xmin(1)     
        magList%fluxConstantError = fit%constantError     
       
        
    end if

!write(*, *) "Computed fluxes=", fit%flux

!call fatal("stopped at line 457")
end subroutine FitSpectra
!********************************************************
!
!*******************************************************
logical(boolean) Function FitSpectrum(x, m, redChiSq, residual)
    real(double) :: x(:)
    integer, intent(in) :: m
    real(double), intent(out) :: redChiSq
    real(single), intent(out) :: residual(:)
   
    real(double) :: chisQ
    integer :: i
    real(double) :: f(7)
    integer :: noOfDegreesOfFreedom
 

!    write(errstr, '(a, i2, a, i2)') "Fitting model to template spectrum ", number, " m=", m
!    call message(AppMsg, Verbose)
    
     
    f = 0.0_double

 
 
    x(1) = ComputeFitConstant(m, f)


    ! Compute chi-squared
    noOfDegreesOfFreedom = m - 1 
   
    residual = real(f)
    chisq = 0.0_double

    do i = 1, m
        chisq = chisq + f(i) * f(i) * weight(i) 
   
    end do


    redchisq = chisq / real(m - 1) 

!write(*, *) "Line 488 redchisq=", redchisq
    if (redchisq .gt. 1.0) then
        fitSpectrum = .false.

!if (count .lt. 50) then 
!write(*, *) "Fit constant = ", x(1)
!    do i = 1, m
        
!      write(*, *) i, "fitdata=", fitData(i), " modelData=", modelData(i), " fit=", (modelData(i) * x(1)), &
!           " residual=", f(i), " weight=", weight(i) 
!    end do

!end if

        
    else
        fitSpectrum = .true.


        
    end if


end Function FitSpectrum


!*******************************************
!
!*****************************************
real(double) Function ComputeFitConstant(m, f)
    integer, intent(in) :: m
    real(double) :: f(:) 
    integer :: i
    real(double) :: sum1, sum2 


    sum1 = 0.0_double
    sum2 = 0.0_double

!weight = 1.0

    
    do i = 1, m
        sum1 = sum1 + weight(i) * modelData(i) * fitdata(i)
        sum2 = sum2 + weight(i) * modelData(i) * modelData(i)
      
    
    end do

    ComputeFitConstant = sum1 / sum2  

!   if (count .lt. 10) then
!       write(*, *) "ComputeFitConstant=", ComputeFitConstant
!       count = count + 1
!   end if

    do i = 1, m
        f(i) = fitData(i) - ComputeFitConstant * modelData(i) 
!if (count .lt. 10) then

!write(*, *) i, " modeldata=", modelData(i), " fitData=", fitData(i), " f=", f(i), " residual=", f(i)
!end if
    end do



end function ComputeFitConstant




!*******************************************
!
!*****************************************
real(double) Function ComputeFitConstantError(a, m)
    real(double), intent(in) :: a
    integer, intent(in) :: m
  
    integer :: i
    real(double) :: sum1, sum2, residual 
  
    
    ! Error in fit constant
    ! Formula used is that from P. Van de kamp, 1967, Principles of Astrometry, W.H. Freeman and Company, page. 217.

    sum1 = 0.0_double
    sum2 = 0.0_double
    do i = 1, m
        residual = fitdata(i) - modeldata(i) * a
        sum1 = sum1 + weight(i) * residual * residual
        sum2 = sum2 + weight(i) * modeldata(i) * modeldata(i)
        

    end do

    ComputeFitconstantError = 0.6745 * sqrt(sum1 / dble(m - 1)) / sqrt(sum2) ! The probable error

if (ComputeFitConstantError .gt. 100.0) then
write(*, *) "ComputeFitConstantError line 581-   ComputeFitconstantError=",  ComputeFitconstantError
write(*, *) "a=", a, " m=", m
  sum1 = 0.0_double
    sum2 = 0.0_double
    do i = 1, m
        residual = fitdata(i) - modeldata(i) * a
write(*, *) i, " fitdata=",  fitdata(i), " model=",  modeldata(i), " residual=", residual, " weight=", weight(i) 
        sum1 = sum1 + weight(i) * residual * residual
        sum2 = sum2 + weight(i) * modeldata(i) * modeldata(i)
        

    end do
end if

 

end function ComputeFitConstantError


!*****************************************************
!
!****************************************************

subroutine updateLogFile(number, rmsResidual, templateNo, noOfDegreesOfFreedom)
    integer, intent(in) :: number
    real(double), intent(in) :: rmsResidual
 
    integer(int16), intent(in) :: templateNo, noOfDegreesOfFreedom

    write(10, '(i12, i10, e10.4, i5)') number, templateNo, rmsResidual, noOfDegreesOfFreedom
       

end subroutine updateLogFile

!*****************************************************
!
!****************************************************
subroutine PlotFit(number, fit)
    integer, intent(in) :: number
    type(fittedData), intent(in) :: fit                                                      
 
   
    character(100) :: title, text
    real(double) :: residual(7), constant
 
    constant = maxval(modelData)
    modelData = modelData / constant
    fitData = fitData / constant
    fitDataErr = fitDataErr / constant


    write(title, '(a, i5, a, i2)') "Fit to source ", number

    call DrawGraph(mFit, real(wData), real(modelData), &
           trim(title), "Wavelength", "Normalized Counts", &
               0.1, 0.9, 0.1, 0.9)

    call pgslw(1)
    call pgsch(0.6)
    call pgpoint(mFit, real(wData), real(fitData), 5)
    call pgerry(mFit, real(wData), real(fitdata - fitDataErr), real(fitData + fitDataErr), 0)

    call pgslw(1)
    call pgsch(0.8)
    write(text, '(a, i3, a, a)')  "Spectrum template number ", fit%templateNo, &
        " Spectral type=", fit%spectralType
    call pgmtxt('T', 2.0, 0.0, 0.0, trim(text))
   

    write(text, '(a, e12.4, a, i2, a, e12.4)') "Scale factor=", &
         fit%constant, " N.D.F.=", fit%ndf, &
        " Reduced chi-squared=",  fit%redchisq
    call pgmtxt('T', 1.0, 0.0, 0.0, trim(text))
    call pgpage


    write(title, '(a, i5, a, i2)') "Fit to source ", number, " Residuals (fit - model)"
   
    call DrawGraph(mFit, real(wData), fit%residual, &
           trim(title), "Wavelength", "Residuals (counts)", &
               0.1, 0.9, 0.1, 0.9)


    call pgslw(1)
    call pgsch(0.6)
    
    call pgerry(mFit, real(wData), real(residual - fitDataErr), real(residual + fitDataErr), 0)

    call pgslw(1)
    call pgsch(0.8)
    write(text, '(a, e12.4, a, i2, a, e12.4)') "Scale factor=", &
         fit%constant, " N.D.F.=", fit%ndf, &
        " Reduced chi-squared=",  fit%redchisq
     

    call pgmtxt('T', 1.0, 0.0, 0.0, trim(text))
    call pgpage

end subroutine PlotFit
!************************************************************
! AddFluxColumn
! Adds a column to the new combined source-list file
! for the computed fluxes
!
!************************************************************
subroutine AddTemplateFluxColumns(n, magTable, raSortIndex, srcTable)
    integer, intent(in) :: n
    type(SourceData), pointer :: magTable(:)
    

    integer, intent(in) :: raSortIndex(:)
    type(TableT), intent(inout) :: srcTable
  
    integer :: i, j, m
    character(20) :: label
    
    integer :: position    
    real(single), pointer :: flux(:) => null()
    real(single), pointer :: temp_table(:, :) => null()
    integer :: nrows, ncols    


    write(errstr, '(a, a)') "Adding flux columns to new source-list file ", trim(stringParameter("outset"))
    call message(AppMsg, Verbose)
    m = magTable(1)%noFilters
  
 
    label = trim(magTable(1)%filter(m)) // "_Flux" 

    if (hasColumn(srcTable, "STD_COLS_ERR")) then    
        position = columnNumber(srcTable, "STD_COLS_ERR") + 1
    else
        position = numberOfColumns(srcTable)
    end if
 
    flux => allocatesingle(1, n, " flux")

!write(*, *) "Line 719"



    do j = 1, 7

!write(*, *) "Line 725 - j=", j

       label = trim(magTable(1)%filter(j)) // "_FLUX"
       do i = 1, n
           flux(i) = magTable(i)%flux(j) 
       end do
       where(flux .lt. 0.0) flux = nullvalue



       call AddReal32IndexColumn(srcTable, label, "ergs/s/cm**2/A", "  ", n, flux, raSortIndex, nullValue, position)


!       position = position + 1
!       label = trim(magTable(1)%filter(j)) // "_COUNTS"
!       do i = 1, n
!           flux(i) = magTable(i)%counts(j)
!       end do  
!       where(flux .lt. 0.0) flux = nullvalue


!       call AddReal32IndexColumn(srcTable, label, "counts/s", "  ", n, flux, raSortIndex, nullValue, position)
    
      position = position + 1

    end do

  

    ! Add the spectra type

    call AddCharIndexColumn(srcTable, "SPEC_TYPE", " ", "Best flux-matched spectral type", n, &
         RASortIndex, magTable%spectralType, position)
 
    position = position + 1

    ! Add thew reduced chi-squared value

    flux = magTable(1 : n)%redchisq 

  

    call AddReal32IndexColumn(srcTable, "RED_CHISQ", " ", "Model fit reduced chi-squared  ", &
           n, flux, raSortIndex, nullValue, position)


   position = position + 1
   flux = magTable(1 : n)%fluxConstant
where(flux .lt. 0.0) flux = nullValue

    call AddReal32IndexColumn(srcTable, "FLUX_CONST", " ", "Model fit flux-constant", &
           n, flux, raSortIndex, nullValue, position)





   position = position + 1
   flux = magTable(1 : n)%fluxConstantError
where(flux .lt. 0.0) flux = nullValue
    call AddReal32IndexColumn(srcTable, "FLUX_C_ERR", " ", "Model fit flux-constant error", &
           n, flux, raSortIndex, nullValue, position)


    position = position + 1


write(*, *) "Line 789"

  

  ! Add the flux tables
  
    temp_table => allocate2dsingle(1, n, 1, 7,  "temp_table")




    do i = 1, n
        temp_table(i, 1 : 7) = magTable(i)%flux(1 : 7)
    end do


 
   call AddVectorColumn(tab=srcTable, data=temp_table, ndata=n, ncols=7, storeindex=RASortIndex, &
            heading="FLUX_TEMPLATE", units="ergs/s/cm**2/A", &
            comment="Computed from fitting template-spectra to source-rates", &
            nullvalue=nullValue, position=position)


write(*, *) "Line 813"


 

 
    ! Add the flux-error tables
    position = position + 1
    do i = 1, n
        temp_table(i, 1 : 7) = magTable(i)%fluxErr(1 : 7)
    end do

    call AddVectorColumn(tab=srcTable, data=temp_table, ndata=n, ncols=7, storeindex=RASortIndex, &
            heading="FLUX_TEMPLATE_ERR", units="ergs/s/cm**2/A", &
            comment=" ", nullvalue=nullValue, position=position)
 


write(*, *) "Line 830"




    deallocate(flux, temp_table)

call LoadRealVectorColumn(srcTable, "FLUX_TEMPLATE", temp_table, nrows, nCols)

 write(*, *) "Line 839"




    nullify(flux, temp_table )
end subroutine AddTemplateFluxColumns
!***************************************************************************
!
!
!**************************************************************************
subroutine AddMeanValTableColumn(tab, data, nData, nCols, colIndex, storeIndex, &
      heading, units, comment, nullValue)
 
    type(tableT), intent(inout) :: tab
    real(single), pointer :: data(:, :)
    integer, intent(in) :: nData, nCols
    integer, intent(in) :: colIndex(:), storeIndex(:)
    character(*), intent(in) :: heading, units, comment
    real(single), intent(in) :: nullValue
    integer :: i, j, k, l
    real(double) :: sum
    integer, dimension(1) :: s 


    s(1) = 1


    do i = 1, nData   
        k = 0
        sum = 0.0_double
        do j = 1, nCols
            l = colIndex(j)
            if (data(i, l) .ne. nullValue) then
                sum = sum + data(i, l)
                k = k + 1
            end if
        end do                       
        if (k .gt. 0) then
            if (k .eq. 1) then
                data(i, 1) = sum
            else
                data(i, 1) = sum / dble(k)
            end if  
        end if
    end do
!write(*, *) "Line 831 AddMeanValTableColumn minval data(:, 1) =", minval(data(:, 1)), " max=", maxval(data(:, 1))
!write(*, *) "heading = ", heading
    call AddReal32IndexColumn(tab, heading, units, comment, nData, data(:, 1), storeIndex, nullValue)
 
!write(*, *) "Line 835 AddMeanValTableColumn minval data(:, 1) =", minval(data(:, 1)), " max=", maxval(data(:, 1))
 

end subroutine AddMeanValTableColumn






end module fluxfit_mod









